def activity_batch_complete(session, actividades, resultado, justification):

    acts_dict = []

    for act in actividades:
        if act.status == "PENDING":
            act_dict = {"_attributes": {}, "_links": {"workitem": {}}}
            act_dict["_attributes"]["type"] = act.type
            act_dict["_attributes"]["name"] = act.name
            act_dict["_links"]["workitem"]["href"] = act.workitem_href

            acts_dict.append(act_dict)

    r = session.restclient.completarActividades(acts_dict, resultado, justification)

    return r



import json
import requests
import urllib
from urllib.parse import urlencode
from pyisim.exceptions import *

requests.packages.urllib3.disable_warnings()

# cat=organizationunits/bporganizations
# si atributos="*" devuelve todos los atributos
# el filtro no importan mayus y es va con comodines (*filtro*)
# si buscar_igual=True entonces quita comodines de busqueda
# si viene con algún atrubuto, lo recupera


class ISIMClient:
    def __init__(self, url, user_, pass_, cert_path=None):

        self.__addr = url
        self.s, self.CSRF = self.autenticar(user_, pass_, cert_path)

    def autenticar(self, user_, pass_, cert=None):

        assert cert is not None, "No certificate passed"
        url = self.__addr + "/itim/restlogin/login.jsp"
        s = requests.Session()
        # print(cert)
        s.verify = cert
        headers = {"Accept": "*/*"}
        r1 = s.get(url, headers=headers)

        assert 404 != r1.status_code, "Error 404: " + r1.text
        # jsessionid=self.s.cookies.get("JSESSIONID")

        url = self.__addr + "/itim/j_security_check"
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        # cookies={"j_username_tmp":user_,"j_password_tmp":pass_}
        data_login = {"j_username": user_, "j_password": pass_}
        r2 = s.post(url, headers=headers, data=data_login)

        # print(r2)
        url = self.__addr + "/itim/rest/systemusers/me"
        r3 = s.get(url, headers=headers)
        try:
            CSRF = r3.headers["CSRFToken"]
        except KeyError:
            # print(user_, pass_)
            raise AuthenticationError(
                "Error de autenticación, verifique sus credenciales."
            )
        return s, CSRF

    def buscarOUs(self, profile_name, filtro, buscar_por=None, attributes="", limit=100):

        url = self.__addr + "/itim/rest/organizationcontainers/" + profile_name
        tipos = ["bporganizations", "organizationunits", "organizations","locations","admindomains"]
        if profile_name not in tipos:
            raise Exception(
                "No es una categoría de OU válida. Seleccione un tipo de categoría entre las siguientes: "
                + str(tipos)
            )

        name_attrs={
            "bporganizations":"ou",
            "organizationunits":"ou",
            "organizations":"o",
            "locations":"l",
            "admindomains":"ou",
        }

        if not buscar_por:
            buscar_por=name_attrs[profile_name]
            
        data = {"attributes": attributes,"limit":limit,buscar_por:filtro}

        OUs = json.loads(self.s.get(url, params=data).text)

        return list(OUs)

    # si filtro="*" busca todo
    def buscarPersonas(
        self, perfil, atributos="cn", embedded="", buscar_por="cn", filtro="*", limit=50
    ):

        assert perfil.lower() in ("person", "bpperson")

        url = self.__addr + "/itim/rest/people"
        if perfil.lower() == "bpperson":
            url = url + "/bpperson"

        data = {
            "attributes": atributos,
            "embedded": embedded,
            "limit": limit,
            buscar_por: filtro,
        }
        headers={
            "Cache-Control":"no-cache"
        }
        data = urlencode(data, quote_via=urllib.parse.quote)

        try:
            response = self.s.get(url, params=data,headers=headers).text
            if response.find("ISIMLoginRequired") != -1:
                # TODO handle this
                raise Exception("Please login.")
            personas = json.loads(response)
        except Exception as e:
            personas = []

        return list(personas)

    def crearPersona(self, person, orgid, justification):

        url = self.__addr + "/itim/rest/people"

        # tipo_persona = person.profile_name

        # tipos = ["Person", "BPPerson"]
        # if tipo_persona in tipos:
        #     raise Exception(
        #         "No es una tipo válido de persona. Seleccione un tipo de persona entre los siguientes: "+str(tipos))

        person_data = person.__dict__.copy()
        person_data.pop("changes", "")

        data = {
            "justification": justification,
            "profileName": person.profile_name,
            "orgID": orgid,
            "_attributes": person_data,
        }

        headers = {
            "CSRFToken": self.CSRF,
            "Content-Type": "application/json",
            "Accept": "*/*",
            # "X-HTTP-Method-Override": "submit-in-batch" FP2
        }

        ret = self.s.post(url, json=data, headers=headers)
        return json.loads(ret.text)

    def modificarPersona(self, href, changes, justification):
        url = self.__addr + href

        data = {
            "justification": justification,
            "_attributes": changes,
        }

        headers = {
            "CSRFToken": self.CSRF,
            "Content-Type": "application/json",
            "Accept": "*/*",
        }

        ret = self.s.put(url, json=data, headers=headers)
        return json.loads(ret.text)

    def buscarAcceso(
        self,
        by="accessName",
        atributos="accessName",
        filtro="*",
        limit=20,
        requestee_href=None,
    ):

        url = self.__addr + "/itim/rest/access"

        data = {
            by: filtro,
            "attributes": atributos,
            "limit": limit,
            "requestee": requestee_href,
            # "filterId": "accessSearch"
        }
        data = urlencode(data, quote_via=urllib.parse.quote)

        res = self.s.get(url, params=data).text
        accesos = json.loads(res)

        return list(accesos)

    def verificarResultadoUnico(self, json_):
        if len(json_) > 1:
            raise MultipleFoundError()
        elif len(json_) == 0:
            raise NotFoundError()
        else:
            return json_[0]

    def obtenerLinks(self, json_, tipoObjeto):

        tipos = {"acceso": "access", "persona": "self"}
        assert tipoObjeto in tipos.keys()
        tipo = tipos[tipoObjeto]

        json_ = self.verificarResultadoUnico(json_)

        return {"_links": {tipo: {"href": json_["_links"]["self"]["href"]}}}

    def buscarActividad(
        self, solicitudID="", search_attr="activityName", search_filter="*"
    ):

        url = self.__addr + "/itim/rest/activities"
        data = {
            "filterId": "activityFilter",
            "status": "PENDING",
            search_attr: search_filter,
        }

        headers = {"Cache-Control": "no-cache"}

        actividades = json.loads(self.s.get(url, params=data, headers=headers).text)
        if solicitudID:
            actividades = filter(
                lambda a: a["_links"]["request"]["href"] == solicitudID, actividades
            )
            return list(actividades)

        return list(actividades)

    def solicitarAccesos(self, accesos, persona, justification):
        url = self.__addr + "/itim/rest/access/assignments"

        persona_rest = {"self": {"href": persona.href}}

        accesos_rest = [{"_links": {"access": {"href": a.href}}} for a in accesos]

        headers = {
            "CSRFToken": self.CSRF,
            "Content-Type": "application/json",
            "Accept": "*/*",
            "X-HTTP-Method-Override": "submit-in-batch",
        }

        # persona: {'self': {'href': '/itim/rest/people/ZX...bnNpb25lcw'}}
        # accesos: [{'_links':{'access': {'href': '/itim/rest/access/32...7126878451'}}},...]
        data = {
            "justification": justification,
            "requests": [
                {
                    "requestee": {
                        "_links": persona_rest,
                        "add": {"assignments": accesos_rest},
                    }
                }
            ],
        }

        # print(data)
        return self.s.post(url, json=data, headers=headers)

    def parse_rfi_form(self, workitem_id, rfi_values):

        response = self.s.get(
            f"{self.__addr}/itim/rest/activities/rfiformdetails/{workitem_id}"
        )
        form_details = json.loads(response.text)
        # esto es un arreglo con la info del formulario
        form = form_details["template"]["page"]["body"]["tabbedForm"]["tab"]

        rfi_form = []
        for tab in form:
            for element in tab["formElement"]:
                attr_name = element["name"].split(".")[-1]
                editable = element["editable"]
                value = ""

                try:
                    required = element["required"]
                except KeyError:
                    required = False

                if required:
                    try:
                        value = form_details["defaultAttrValues"][attr_name]
                    except KeyError:
                        pass
                if editable:
                    value = [
                        attr["value"]
                        for attr in rfi_values
                        if attr["name"] == attr_name
                    ][0]

                if editable or required:
                    rfi_form.append(
                        {
                            "name": attr_name,
                            "value": value,
                        }
                    )

        return rfi_form

    # falta tener en cuenta cuando llegan varias actividades
    def completarActividades(self, actividades, resultado, justification="ok"):

        url = self.__addr + "/itim/rest/workitems"

        resultCodes = {
            "approve": "AA",
            "reject": "AR",
            "successful": "SS",
            "warning": "SW",
            "failure": "SF",
        }

        body = []
        if len(actividades) == 0:
            raise Exception("Nothing to complete")

        for activity in actividades:
            activityType = activity["_attributes"]["type"]
            activityLabel = activity["_attributes"]["name"]
            workitem = activity["_links"]["workitem"]["href"]

            if activityType == "APPROVAL":
                assert resultado in ["approve", "reject"]
            elif activityType == "WORK_ORDER":
                assert resultado in ["successful", "warning", "failure"]
            elif activityType == "RFI":
                assert isinstance(resultado, list)

            resultCode = resultCodes[resultado] if activityType != "RFI" else "RS"
            action = {
                "_links": {"self": {"href": workitem}},
                "action": {"code": resultCode},
                "label": activityLabel,
                "justification": justification,
            }

            if activityType == "RFI":

                assert (
                    len(actividades) == 1
                ), "Solo es posible completar un RFI a la vez"

                workitem_id = workitem.split("/")[-1]

                action = {
                    "action": {"code": resultCode},
                    "label": activityLabel,
                    "justification": justification,
                }

                if len(resultado) > 0:

                    rfi_form = self.parse_rfi_form(workitem_id, resultado)
                    action["rfiAttributeValues"] = rfi_form

                headers = {
                    "CSRFToken": self.CSRF,
                    "Content-Type": "application/json",
                    "Accept": "*/*",
                }

                return self.s.put(f"{url}/{workitem_id}", json=action, headers=headers)

            body.append(action)

        headers = {
            "CSRFToken": self.CSRF,
            "Content-Type": "application/json",
            "Accept": "*/*",
            "X-HTTP-Method-Override": "submit-in-batch",
            "methodOverride": "submit-in-batch",
        }

        return self.s.put(url, json=body, headers=headers)

    def buscarFormulario(self, perfil):

        url = self.__addr + "/itim/rest/forms/people"

        assert perfil in ["Person", "BPPerson"], "Perfil inválido."

        urlPerfil = url + "/" + perfil
        resp = self.s.get(urlPerfil)

        return json.loads(resp.text)["template"]["page"]["body"]["tabbedForm"]["tab"]

    def buscarServicio(self, search_attr, search_filter, limit, atributos=""):

        url = self.__addr + "/itim/rest/services"

        data = {
            search_attr: search_filter,
            "attributes": atributos,
            "limit": limit,
        }
        data = urlencode(data, quote_via=urllib.parse.quote)

        servicios = json.loads(self.s.get(url, params=data).content)

        if len(servicios) == 0:
            raise NotFoundError(
                f"Servicio no encontrado: ({search_attr}={search_filter})"
            )

        return servicios

    def eliminarServicio(self, nombre):

        url = self.__addr + "/itim/rest/services"

        servicio = self.buscarServicio(nombre)
        servicio_href = servicio[0]["_links"]["self"]["href"]
        servicio_id = servicio_href.split("/")[-1]

        url_del = f"{url}/{servicio_id}"

        headers = {
            "CSRFToken": self.CSRF,
            "Content-Type": "application/json",
            "Accept": "*/*",
            "X-HTTP-Method-Override": "submit-in-batch",
        }

        return self.s.delete(url_del, headers=headers)

    def lookupSolicitud(self, requestID):
        url = self.__addr + "/itim/rest/requests"

        url_req = url + "/" + requestID
        data = {"attributes": "*"}

        solicitud = json.loads(self.s.get(url_req, params=data).text)

        return solicitud

    def lookupActividad(self, activityID):
        url = self.__addr + "/itim/rest/activities"

        url_act = url + "/" + activityID
        data = {"attributes": "*"}

        actividad = json.loads(self.s.get(url_act, params=data).text)

        return actividad

    def lookupPersona(self, href,attributes="dn"):
        url = self.__addr + href

        params = {
            "attributes": attributes,
            "forms": False,
        }

        person = self.s.get(url, params=params)

        return json.loads(person.text)

    def lookupCurrentPerson(self,attributes="*",embedded=""):
        url = self.__addr + "/itim/rest/people/me"

        params = {
            "attributes": attributes,
            "embedded": embedded,
        }

        person = self.s.get(url, params=params)

        return json.loads(person.text)




import datetime
from collections import defaultdict

from pyisim.exceptions import *


class ProvisioningPolicy:
    """
    Para usar con el API SOAPWS de ISIM7
    """

    def __init__(
        self,
        session,
        provisioning_policy=None,
        policy_attrs=None,
    ):
        """Initializes ProvisioningPolicy object

        Args:
            session (pyisim.Session): Custom ISIM/ Default ISIM Session
            policy_attrs(dict):
                name (str): policy name
                description (str): policy description
                entitlements (dict): {
                    service_dn/service_profile_name/*:{
                        "automatic":True/False,
                        "workflow":account_request_workflow_name/None,
                        "parameters:{
                            attr_name:[
                                    {
                                        "enforcement":enforcement_type (allowed/mandatory/default/excluded),
                                        "type":parameter_type (constant/regex/null/script)
                                        "values": attribute_value (None, str for regex/script, list for constant values),
                                    },
                                    {... more values}
                                ],
                            {... more attributes}
                        }
                    },
                    {... more services}
                }
                memberships (list): list of role DNs or ["*"] if everyone. TYPE 4 NOT SUPPORTED (Everyone that's not entitled through other policies NOT SUPPORTED)
                parent (pyisim.entities.OrganizationalContainer): parent container
                priority (int): policy priority.
                scope (int, optional): policy scope (1=ONE_LEVEL / 2=SUBTREE). Defaults to 2.
                enabled (bool, optional): Defaults to True.
                caption(str,optional): Defaults to "".
                keywords(str,optional): Defaults to "".
        """

        # session = session.soapclient
        url = session.soapclient.addr + "WSProvisioningPolicyServiceService?wsdl"
        self.__pp_client = session.soapclient.get_client(url)

        local_tz = datetime.datetime.now().astimezone().tzinfo
        self.date = datetime.datetime.now(local_tz).isoformat()

        if policy_attrs:
            self.description = policy_attrs["description"]
            self.name = policy_attrs["name"]
            self.ou = policy_attrs["parent"]  # .wsou
            self.entitlements = policy_attrs["entitlements"]
            self.membership = policy_attrs["memberships"]

            # probando
            # self.entitlements = self.__crearEntitlementList(
            #     session.soapclient, policy_attrs["entitlements"]
            # )
            # self.membership = self.__crearMembershipList(
            #     session.soapclient, policy_attrs["memberships"]
            # )
            self.priority = policy_attrs["priority"]
            self.scope = (
                policy_attrs["scope"] if "scope" in policy_attrs else 2
            )  # 1=ONE_LEVEL, 2=SUBTREE
            self.enabled = (
                policy_attrs["enabled"] if "enabled" in policy_attrs else True
            )
            self.caption = policy_attrs["caption"] if "caption" in policy_attrs else ""
            self.keywords = (
                policy_attrs["keywords"] if "keywords" in policy_attrs else ""
            )

        else:

            # if dn: no hay método de lookup en sim
            #     rol=session.lookupProvisioningPolicy(dn)
            self.description = provisioning_policy["description"]
            self.name = provisioning_policy["name"]
            self.dn = provisioning_policy["itimDN"]
            self.ou = OrganizationalContainer(
                session, dn=provisioning_policy["organizationalContainer"]["itimDN"]
            )
            self.priority = provisioning_policy["priority"]
            self.scope = provisioning_policy["scope"]

            # self.entitlements = provisioning_policy["entitlements"]
            # for titularidad in self.entitlements.item:
            #     if titularidad.parameters.parameters is None:
            #         titularidad.parameters.parameters = {"item": []}

            self.entitlements = self.__traducirWSEntitlements(
                session.soapclient, provisioning_policy["entitlements"]
            )
            self.membership = [
                m["name"] for m in provisioning_policy["membership"]["item"]
            ]

            self.caption = provisioning_policy["caption"]
            self.keywords = provisioning_policy["keywords"]
            self.enabled = provisioning_policy["enabled"]

    def __traducirWSEntitlements(self, session, wsentitlements):
        """
        Convierte WSEntitlements en diccionario estándar para poder modificar
        Retorna:
        entitlements (dict): {
            service_dn/service_profile_name/*:{
                "automatic":True/False,
                "workflow":account_request_workflow_name/None,
                "parameters:[
                    attr_name:[
                            {
                                "enforcement":enforcement_type (allowed/mandatory/default/excluded),
                                "type":parameter_type (constant/regex/null/script)
                                "values": attribute_value (None, str for regex/script, list for constant values),
                            },
                            {... more values}
                        ],
                    {... more attributes}
                ]
            },
            {... more services}
        }
        """

        enforcement_mapping = {
            2: "default",
            1: "allowed",
            0: "excluded",
            3: "mandatory",
        }

        type_mapping = {
            0: "constant",
            10: "script",
            20: "regex",
        }

        entitlements = {}
        for titularidad in wsentitlements["item"]:
            service = titularidad["serviceTarget"]["name"]
            workflow = titularidad["processDN"]
            automatic = titularidad["type"] == 1
            wsparams = titularidad["parameters"]["parameters"]
            if not wsparams:
                parameters = {}
            else:
                parameters = {}
                for param in wsparams["item"]:
                    attr_name = param.name
                    values = param.values.item
                    types = param.expressionTypes.item
                    enforcement = param.enforcementTypes.item
                    attr_values = list(zip(values, enforcement, types))
                    attr_values = [
                        {
                            "enforcement": enforcement_mapping[val[1]],
                            "type": "constant"
                            if val[0][0] == val[0][-1] == '"'
                            else "null"
                            if val[0] == "return null;"
                            else type_mapping[val[2]],
                            "values": [val[0][1:-1]]
                            if val[0][0] == val[0][-1] == '"'
                            else None
                            if val[0] == "return null;"
                            else val[0],
                        }
                        for val in attr_values
                    ]

                    parameters[attr_name] = attr_values

            entitlements[service] = {
                "workflow": workflow,
                "automatic": automatic,
                "parameters": parameters,
            }

        return entitlements

    def __crearEntitlementList(self, session, titularidades):
        """
        {
            service_dn/service_profile_name/*:{
                "automatic":True/False,
                "workflow":account_request_workflow_name/None,
                "parameters:[
                    attr_name:[
                            {
                                "enforcement":enforcement_type (allowed/mandatory/default/excluded),
                                "type":parameter_type (constant/regex/null/script)
                                "values": attribute_value (None, str: for regex/script, list: for constant value),
                            },
                            {... more values}
                        ],
                    {... more attributes}
                ]
            },
            {... more services}
        }
        """
        # session=session.soapclient
        client = self.__pp_client

        itemFactory = client.type_factory("ns1")
        listFactory = client.type_factory("ns0")

        entitlement_list = []

        for servicio, attrs in titularidades.items():

            if servicio == "*":  # Todos los servicios
                servicio_dn = "*"
                tipo_entitlement = 2

            elif "profile" in servicio.lower():  # Todos los servicios de un perfil
                # servicio_dn=session.buscarPerfilServicio(servicio) no sirve búsqueda de perfiles
                servicio_dn = servicio
                tipo_entitlement = 0

            else:  # Servicio específico
                servicio_dn = servicio
                tipo_entitlement = 1

            serviceTarget = itemFactory.WSServiceTarget(
                name=servicio_dn, type=tipo_entitlement
            )

            type_ = 1 if attrs["automatic"] else 0
            process_dn = (
                session.searchWorkflow(attrs["workflow"], self.ou.name)
                if attrs["workflow"]
                else None
            )
            # print(process_dn)

            parameters = self.__crearParameterList(client, attrs["parameters"])

            wsentitlement = itemFactory.WSProvisioningPolicyEntitlement(
                ownershipType="Individual",
                type=type_,
                serviceTarget=serviceTarget,
                parameters=parameters,
                processDN=process_dn,
            )
            entitlement_list.append(wsentitlement)

        ws_entitlement_list = listFactory.ArrayOf_tns1_WSProvisioningPolicyEntitlement(
            entitlement_list
        )

        return ws_entitlement_list

    def __crearParameterList(self, client, atributos):
        """
        {
             attr_name:[
                     {
                         "enforcement":enforcement_type (allowed/mandatory/default/excluded),
                         "type":parameter_type (constant/regex/null/script)
                         "values": attribute_value (None, str: for regex/script, list: for constant value),
                     },
                     {... more values}
                 ],
             {... more attributes}
         }
        """
        itemFactory = client.type_factory("ns1")
        listFactory = client.type_factory("ns0")

        attrs = atributos.copy()

        parameters = []

        enforcement_map = {
            "default": 2,
            "allowed": 1,
            "excluded": 0,
            "mandatory": 3,
            "optional": 2,
        }

        type_map = {
            "constant": 0,
            "script": 10,
            "regex": 20,
            "null": 10,
        }

        for name, values in attrs.items():
            attr_values = []
            for value in values:
                if "constant" in value["type"].lower():
                    val = [f'"{s}"' for s in value["values"]]
                    enforce = [enforcement_map[value["enforcement"].lower()]] * len(val)
                    tipo = [0] * len(val)
                elif "null" in value["type"].lower():
                    val = ["return null;"]
                    enforce = [enforcement_map[value["enforcement"].lower()]]
                    tipo = [10]
                else:
                    val = [value["values"]]
                    enforce = [enforcement_map[value["enforcement"].lower()]]
                    tipo = [type_map[value["type"].lower()]]

                attr_values.append((val, enforce, tipo))

            # organiza la info al formato requerido (listas planas separadas)
            values, enforcements, types = list(zip(*attr_values))

            values = [item for elem in values for item in elem]
            enforcements = [item for elem in enforcements for item in elem]
            types = [item for elem in types for item in elem]

            values = listFactory.ArrayOf_xsd_string(values)
            expressionTypes = listFactory.ArrayOf_xsd_int(types)
            enforcementTypes = listFactory.ArrayOf_xsd_int(enforcements)

            param = itemFactory.WSServiceAttributeParameter(
                name=name,
                values=values,
                expressionTypes=expressionTypes,
                enforcementTypes=enforcementTypes,
            )
            parameters.append(param)

        parameter_array = listFactory.ArrayOf_tns1_WSServiceAttributeParameter(
            parameters
        )
        wsparameters = itemFactory.WSProvisioningParameters(parameter_array)

        return wsparameters

    def __crearMembershipList(self, session, memberships):
        """
        Si recibe ["*"]: Devuelve 2;*
        Si recibe lista con nombres de roles: Devuelve array[1;rol1, 2;rol2, ...]
        """
        # session=session.soapclient
        client = self.__pp_client

        membershipFactory = client.type_factory("ns1")
        listFactory = client.type_factory("ns0")

        if not memberships:
            raise Exception("La política debe tener al menos un miembro.")
        elif memberships[0] == "*":
            mem = membershipFactory.WSProvisioningPolicyMembership(name="*", type=2)
            membershipList = listFactory.ArrayOf_tns1_WSProvisioningPolicyMembership(
                [mem]
            )
        else:
            lista = []
            for membership in memberships:
                role_dn = membership
                mem = membershipFactory.WSProvisioningPolicyMembership(
                    name=role_dn, type=3
                )
                lista.append(mem)
            membershipList = listFactory.ArrayOf_tns1_WSProvisioningPolicyMembership(
                lista
            )

        return membershipList

    def add(self, session):
        session = session.soapclient
        client = self.__pp_client

        itemFactory = client.type_factory("ns1")

        ents = self.__crearEntitlementList(session, self.entitlements)
        membs = self.__crearMembershipList(session, self.membership)
        wspp = itemFactory.WSProvisioningPolicy(
            description=self.description,
            name=self.name,
            entitlements=ents,
            membership=membs,
            priority=self.priority,
            scope=self.scope,
            enabled=self.enabled,
        )

        del wspp["organizationalContainer"]
        del wspp["itimDN"]

        r = session.crearPolitica(self.ou.wsou, wspp, self.date)

        # la respuesta no envía el DN, entonces no se puede meter de una
        return r

    def modify(self, session, changes={}):
        session = session.soapclient
        client = self.__pp_client

        itemFactory = client.type_factory("ns1")

        for attr,value in changes.items():
            setattr(self,attr,value)

        ents = self.__crearEntitlementList(session, self.entitlements)
        membs = self.__crearMembershipList(session, self.membership)

        wspp = itemFactory.WSProvisioningPolicy(
            description=self.description,
            name=self.name,
            entitlements=ents,
            membership=membs,
            priority=self.priority,
            scope=self.scope,
            enabled=self.enabled,
            itimDN=self.dn,
            caption=self.caption,
            keywords=self.keywords,
        )

        del wspp["organizationalContainer"]

        r = session.modificarPolitica(self.ou.wsou, wspp, self.date)

        return r

    def delete(self, session):
        session = session.soapclient
        r = session.eliminarPolitica(self.ou.wsou, self.dn, self.date)
        return r


class StaticRole:
    """
    Para usar con el API SOAPWS de ISIM7
    Los atributos se llenan correctamente únicamente al crear un nueva instancia de StaticRole
    Para enviar al API:
        Instanciar con atributos necesarios
        llamar el método crearWSRole()
        llamar simsoap.crearRolEstatico() con lo retornado del método

    access_option:
        1: No acceso
        2: Acceso
        3: Acceso común

    classification:
        Empresarial
        Aplicacion

    access_category: El nombre de la categoría del acceso del rol tal como se ve en la consola de administración

    owner roles: Lista de nombres de roles dueños del acceso
    owner_cedulas: Lista de cédulas de personas dueñas del acceso

    """

    def __init__(
        self,
        session,
        dn=None,
        rol=None,
        role_attrs=None,
    ):
        """

        Args:
            session (pyisim.Session): session object
            id (str): for role lookup
            rol (zeep.WSRole): for initialization after search
            role_attrs (dict):      name,
                                    description,
                                    parent (pyisim.entities.OrganizationalContainer),
                                    classification: str (eg. role.classification.business). Defaults to "".
                                    access_option (int): 1: disable / 2: enable / 3: shared access
                                    access_category: access category. Defaults to None.
                                    owners: list of owner DNs. can be people or roles. Defaults to [].
        """
        # session = session.soapclient

        url = session.soapclient.addr + "WSRoleServiceService?wsdl"

        self.__role_client = session.soapclient.get_client(url)

        if role_attrs:

            self.name = role_attrs["name"]
            self.description = role_attrs["description"]
            self.ou = role_attrs["parent"]
            self.classification = (
                role_attrs["classification"] if role_attrs["classification"] else ""
            )

            role_attrs["access_option"]=str(role_attrs["access_option"])
            if role_attrs["access_option"] not in ["1", "2", "3"]:
                raise ValueError(
                    "Access option must be an int. 1: disable / 2: enable / 3: shared access"
                )
            self.access_option = role_attrs["access_option"]

            if role_attrs["access_option"] in ["2", "3"]:
                if not role_attrs["access_category"]:
                    raise ValueError("Si el rol es acceso, debe darle una categoría")
                self.access_category = role_attrs["access_category"]

            self.owners = role_attrs["owners"] if role_attrs["owners"] else []

        else:
            if dn:
                rol = session.soapclient.lookupRole(dn)

            self.name = rol["name"]
            self.description = rol["description"]
            self.dn = rol["itimDN"]

            attrs = {
                attr["name"]: [i for i in attr["values"]["item"]]
                for attr in rol["attributes"]["item"]
            }
            attrs = defaultdict(list, attrs)
            ou = attrs["erparent"][0]  # dn del contenedor
            self.parent=OrganizationalContainer(session,dn=ou)
            if attrs["erroleclassification"]:
                self.classification = attrs["erroleclassification"][0]
            if attrs["eraccessoption"]:
                self.access_option = attrs["eraccessoption"][0]
            if attrs["erobjectprofilename"]:
                self.access_category = attrs["erobjectprofilename"][0]
            self.owners = attrs["owner"]

    def crearAtributoRol(self, client, name, values):
        itemFactory = client.type_factory("ns1")
        listFactory = client.type_factory("ns0")

        values = listFactory.ArrayOf_xsd_string(values)

        attr = itemFactory.WSAttribute(
            name=name, operation=0, values=values, isEncoded=False
        )

        return attr

    def crearWSRole(self, session):

        """
        A partir de la información de la instancia, devuelve un objeto WSRole para entregárselo al API de ISIM
        """
        # session=session.soapclient
        client = self.__role_client

        itemFactory = client.type_factory("ns1")
        listFactory = client.type_factory("ns0")

        eraccesoption = self.crearAtributoRol(
            client, "eraccessoption", self.access_option
        )
        erRoleClassification = self.crearAtributoRol(
            client, "erRoleClassification", self.classification
        )

        if self.access_option in ["2", "3"]:
            erObjectProfileName = self.crearAtributoRol(
                client, "erObjectProfileName", self.access_category
            )
            lista_atributos = [eraccesoption, erRoleClassification, erObjectProfileName]
        else:
            lista_atributos = [eraccesoption, erRoleClassification]

        if self.owners is not None:
            owner = self.crearAtributoRol(client, "owner", self.owners)
            lista_atributos.append(owner)

        array_attributes = listFactory.ArrayOf_tns1_WSAttribute(lista_atributos)

        role = itemFactory.WSRole(
            attributes=array_attributes,
            select=False,
            name=self.name,
            description=self.description,
        )

        del role["itimDN"]
        # del role["erparent"]
        return role

    def add(self, session):
        session = session.soapclient

        # client = self.__role_client

        wsrole = self.crearWSRole(session)

        r = session.crearRolEstatico(wsrole, self.ou.wsou)

        self.dn = r["itimDN"]
        return r

    def modify(self, session,changes={}):
        session = session.soapclient
        # url = session.addr + "WSRoleServiceService?wsdl"
        client = self.__role_client

        for attr,value in changes.items():
            setattr(self,attr,value)

        wsrole = self.crearWSRole(session)
        wsattributes = wsrole["attributes"]["item"]

        errolename = self.crearAtributoRol(client, "errolename", self.name)
        description = self.crearAtributoRol(
            client, "description", wsrole["description"]
        )

        wsattributes.append(errolename)
        wsattributes.append(description)

        # print(wsattributes)

        r = session.modificarRolEstatico(self.dn, wsattributes)

        return r

    def delete(self, session, date=None):
        if date:
            raise NotImplementedError(
                "No se ha implementado la programación de tareas."
            )

        r = session.soapclient.eliminarRolEstatico(self.dn, date)

        return r


class OrganizationalContainer:
    def __init__(self, session, dn=None, organizational_container=None):
        if dn:
            self.wsou = session.soapclient.lookupContainer(dn)
            self.name = self.wsou.name
            self.dn = self.wsou["itimDN"]
            self.profile_name = self.wsou["profileName"]

            rest_profile_names={
                "BusinessPartnerOrganization":"bporganizations",
                "OrganizationalUnit":"organizationunits",
                "Organization":"organizations",
                "Location":"locations",
                "AdminDomain":"admindomains",
            }
            self.href=session.restclient.buscarOUs(rest_profile_names[self.profile_name],self.name)[0]["_links"]["self"]["href"]

        elif organizational_container:
            self.name = organizational_container["_links"]["self"]["title"]
            self.href = organizational_container["_links"]["self"]["href"]
            self.dn = organizational_container["_attributes"]["dn"]
            self.wsou = session.soapclient.lookupContainer(self.dn)
            self.profile_name = self.wsou["profileName"]

    def __eq__(self, o) -> bool:
        return self.dn==o.dn


class Person:

    profile_name = "Person"

    def __init__(self, session, person=None, href=None, person_attrs=None):

        self.changes = {}

        if person:
            self.href = person["_links"]["self"]["href"]
            self.dn = session.restclient.lookupPersona(self.href, attributes="dn")[
                "_attributes"
            ]["dn"]
            person_attrs = person["_attributes"]

        elif href:
            r = session.restclient.lookupPersona(href, attributes="*")
            assert (
                r["_links"]["self"]["href"] == href
            ), "Persona no encontrada o inválida"

            self.href = href
            self.dn = session.restclient.lookupPersona(self.href, attributes="dn")[
                "_attributes"
            ]["dn"]
            person_attrs = r["_attributes"]

        for k, v in person_attrs.items():
            setattr(self, k, v)

    def __setattr__(self, attr, val):
        if hasattr(self, attr):
            self.changes[attr] = val
        super().__setattr__(attr, val)

    def __init_subclass__(cls):

        try:
            getattr(cls, "profile_name")
        except AttributeError:
            raise TypeError(
                f"All classes based on the Person entity need their profile specified (Person/BPPerson/Your custom profile) in the profile_name attribute"
            )

        return super().__init_subclass__()

    def add(self, session, parent: OrganizationalContainer, justification):
        orgid = parent.href.split("/")[-1]
        ret = session.restclient.crearPersona(self, orgid, justification)
        return ret

    def modify(self, session, justification, changes={}):
        try:
            href = self.href
            
            self.changes.update(changes)
            # for attr,value in changes.items():
            #     setattr(self,attr,value)

            ret = session.restclient.modificarPersona(
                self.href, self.changes, justification
            )
            return ret
        except AttributeError:
            raise Exception(
                "Person has no reference to ISIM, search for it or initialize it with href to link it."
            )

    def request_access(self, session, accesos, justification):

        ret = {}
        if len(accesos) > 0:
            ret = session.restclient.solicitarAccesos(accesos, self, justification)
        return ret

    def suspend(self, session, justification):

        try:
            try:
                dn = self.dn
            except AttributeError:
                dn = session.restclient.lookupPersona(self.href, attributes="dn")[
                    "_attributes"
                ]["dn"]
                self.dn = dn

            ret = session.soapclient.suspenderPersona(dn, justification)
            return ret
        except AttributeError:
            raise Exception(
                "Person has no reference to ISIM, search for it or initialize it with href to link it."
            )

    def restore(self, session, justification):
        try:
            try:
                dn = self.dn
            except AttributeError:
                dn = session.restclient.lookupPersona(self.href, attributes="dn")[
                    "_attributes"
                ]["dn"]
                self.dn = dn

            ret = session.soapclient.restaurarPersona(self.dn, justification)
            return ret
        except AttributeError:
            raise Exception(
                "Person has no reference to ISIM, search for it or initialize it with href to link it."
            )

    def delete(self, session, justification):

        try:
            try:
                dn = self.dn
            except AttributeError:
                dn = session.restclient.lookupPersona(self.href, attributes="dn")[
                    "_attributes"
                ]["dn"]
                self.dn = dn

            ret = session.soapclient.eliminarPersona(self.dn, justification)
            return ret
        except AttributeError:
            raise Exception(
                "Person has no reference to ISIM, search for it or initialize it with href to link it."
            )


class Activity:
    def __init__(self, session, activity=None, id=None):

        if id:
            activity = session.restclient.lookupActividad(str(id))
            if "_attributes" not in activity.keys():
                raise NotFoundError(f"Activity not found: {id}")

        self.request_href = activity["_links"]["request"]["href"]
        self.href = activity["_links"]["self"]["href"]
        self.workitem_href = activity["_links"]["workitem"]["href"]
        self.name = activity["_attributes"]["name"]
        self.type = activity["_attributes"]["type"]
        self.status = activity["_attributes"]["status"]["key"].split(".")[-1]
        self.requestee = activity["_links"]["requestee"]["title"]

    def complete(self, session, result, justification):

        """Allos to complete:
            Approvals   (result: approve/reject)
            Work Orders (result: successful/warning/failure)
            RFI         (result=[
                            {'name':attr_name,'value':attr_value}, ...
                        ])
        """
        act_dict = {"_attributes": {}, "_links": {"workitem": {}}}
        act_dict["_attributes"]["type"] = self.type
        act_dict["_attributes"]["name"] = self.name
        act_dict["_links"]["workitem"]["href"] = self.workitem_href

        assert self.status == "PENDING", "Activity is already complete."
        r = session.restclient.completarActividades([act_dict], result.lower(), justification)

        return r


class Access:
    def __init__(self, access=None):

        if access:
            self.href = access["_links"]["self"]["href"]
            self.name = access["_links"]["self"]["title"]


class Service:
    def __init__(self, session, service=None):

        # if id:
        #     service=session.restclient.lookupServicio(str(id))
        #     if "_attributes" not in service.keys():
        #         raise NotFoundError(f"Servicio no encontrado {id}")

        if service:
            self.dn = service["itimDN"]
            self.profile_name = service["profileName"]
            self.name = service["name"]


class Group:
    def __init__(self, session, group=None):
        if group:
            self.name = group["name"]
            self.dn = group["itimDN"]
            self.id = group["id"]
            self.description = group["description"]
            self.profile_name = group["profileName"]
            self.attributes = {
                attr.name: [v for v in attr.values.item]
                for attr in group.attributes.item
            }



import pyisim.rest as simrest
import pyisim.soap as simsoap
from pyisim.entities import Person

class Session:
    """
    Maneja sesión conjunta de REST y SOAP
    """

    def __init__(self, url, username, password, certificate_path):
        self.username = username
        self.restclient = simrest.ISIMClient(url, username, password, certificate_path)
        self.soapclient = simsoap.ISIMClient(url, username, password, certificate_path)

    def current_person(self,attributes="*",embedded=""):
        """Returns the current logged in person entity.

        Args:
            attributes (str, optional): comma separated attributes to return. Defaults to "*".
            embedded (str, optional): comma separated embedded attributes to return. Defaults to "".

        Returns:
            pyisim.entities.Person: Person entity of the currently logged user.
        """
        p=self.restclient.lookupCurrentPerson(attributes,embedded)     
        return Person(self,person=p)




import json

def list_files(session):

    url = f"https://{session.base_url}/v1/property/propertyfiles"

    headers = {"Accept": "application/json"}

    r = session.s.get(url, headers=headers, auth=session.auth)
    return json.loads(r.content)


def get_property_value(session, property_file, property_name=None):
    url = f"https://{session.base_url}/v1/property"

    params = {
        "PropertyFile": property_file,
    }
    if property_name:
        params["PropertyName"] = property_name

    headers = {"Accept": "application/json"}

    r = session.s.get(url, params=params, headers=headers, auth=session.auth)

    return json.loads(r.content)


def add_property(session, property_file, property_name, property_value):
    url = f"https://{session.base_url}/v1/property"

    data = {
        "PropertyFile": property_file,
        "PropertyName": property_name,
        "PropertyValue": property_value,
    }

    headers = {
        "Accept": "application/json",
        "Content-Type": "application/json",
    }

    r = session.s.post(url, json=data, headers=headers, auth=session.auth)

    return r.reason


def update_property(session, property_file, property_name, property_value):
    url = f"https://{session.base_url}/v1/property"

    data = {
        "PropertyFile": property_file,
        "PropertyName": property_name,
        "PropertyValue": property_value,
    }

    headers = {
        "Accept": "application/json",
        "Content-Type": "application/json",
    }

    r = session.s.put(url, json=data, headers=headers, auth=session.auth)

    return r.reason


def delete_property(session, property_file, property_name=None):
    url = f"https://{session.base_url}/v1/property"

    params = {
        "PropertyFile": property_file,
    }
    if property_name:
        params["PropertyName"] = property_name

    headers = {
        "Accept": "application/json",
        "Content-Type": "application/json",
    }

    r = session.s.delete(url, params=params, headers=headers, auth=session.auth)

    return r


def create_or_update_property(session, property_file, property_name, property_value):
    old_value = get_property_value(session, property_file, property_name)

    if "result" in old_value.keys() and old_value["result"] == "property_not_found":
        r = "Adding: " + add_property(
            session, property_file, property_name, property_value
        )
    else:
        if property_value == old_value["PropertyValue"]:
            r = "No changes"
        else:
            r = "Updating: " + update_property(
                session, property_file, property_name, property_value
            )

    return r



class AuthenticationError(Exception):
    pass


class NotFoundError(Exception):
    pass


class PersonNotFoundError(NotFoundError):
    pass


class MultipleFoundError(Exception):
    pass


class ContratoNoEncontradoError(Exception):
    pass


class NotImplementedError(Exception):
    pass


class InvalidOptionError(Exception):
    pass



from datetime import date
from zeep import Client, Settings
from zeep.xsd import Nil
from zeep.transports import Transport
from zeep.helpers import serialize_object
from zeep.cache import InMemoryCache

# from isim_classes import StaticRole
import requests
from pyisim.exceptions import *
# from pyisim.entities import OrganizationalContainer


requests.packages.urllib3.disable_warnings()


class ISIMClient:
    def __init__(self, url, user_, pass_, cert_path=None):

        self.addr = url + "/itim/services/"
        self.cert_path = cert_path
        self.s = self.login(user_, pass_)

    def login(self, user_, pass_):
        url = self.addr + "WSSessionService?wsdl"
        assert self.cert_path is not None, "No certificate passed"
        client = self.get_client(url)
        session = client.service.login(user_, pass_)
        return session

    def get_client(self, url):

        # Si ya se inicializó el cliente especificado en client_name, lo devuelve. Si no, lo inicializa, setea y devuelve.
        client_name=url.split("/")[-1][:-5].lower() #ej. -> https://<ITIMURL>/.../WSSessionService?wsdl -> wssessionservice
        client = getattr(self, client_name, None)

        if client is None:
            settings = Settings(strict=False)
            s = requests.Session()
            s.verify = self.cert_path
            client = Client(
                url,
                settings=settings,
                transport=Transport(session=s, cache=InMemoryCache()),
            )
            # necesario porque los WSDL de SIM queman el puerto y no funciona con balanceador
            client.service._binding_options["address"] = url[:-5]
            setattr(self, client_name, client)

        return client

    def lookupContainer(self, dn):

        url = self.addr + "WSOrganizationalContainerServiceService?wsdl"
        client = self.get_client(url)

        cont = client.service.lookupContainer(self.s, dn)

        return cont

    def buscarOrganizacion(self, perfil, nombre):

        url = self.addr + "WSOrganizationalContainerServiceService?wsdl"
        client = self.get_client(url)

        ous = client.service.searchContainerByName(self.s, Nil, perfil, nombre)

        return ous


    def buscarPoliticaSuministro(self, wsou, nombre_politica, find_unique=True):
        """
        Si find_unique es True (por defecto):
           lanza error si no encuentra ninguno o encuentra múltiples.
           Devuelve instancia encontrada

        Si find_unique es False:
           No lanza errores
           Devuelve lista de resultados
        """

        url = self.addr + "WSProvisioningPolicyServiceService?wsdl"
        client = self.get_client(url)

        politicas = client.service.getPolicies(self.s, wsou, nombre_politica)

        if find_unique:
            assert (
                len(politicas) > 0
            ), f"No se ha encontrado la política {nombre_politica}."
            assert (
                len(politicas) == 1
            ), f"Se ha encontrado más de la política con: {nombre_politica}"
            return politicas[0]
        else:
            return politicas

    def crearPolitica(self, ou, wsprovisioningpolicy, date):

        url = self.addr + "WSProvisioningPolicyServiceService?wsdl"
        client = self.get_client(url)

        s = client.service.createPolicy(self.s, ou, wsprovisioningpolicy, date)

        return s

    def modificarPolitica(self, ou, wsprovisioningpolicy, date):

        url = self.addr + "WSProvisioningPolicyServiceService?wsdl"
        client = self.get_client(url)

        s = client.service.modifyPolicy(self.s, ou, wsprovisioningpolicy, date)

        return s

    def eliminarPolitica(self, ou, dn, date):
        url = self.addr + "WSProvisioningPolicyServiceService?wsdl"
        client = self.get_client(url)

        s = client.service.deletePolicy(self.s, ou, dn, date)

        return s

    def buscarRol(self, filtro, find_unique=True):
        """
        Si find_unique es True (por defecto):
           lanza error si no encuentra ninguno o encuentra múltiples.
           Devuelve instancia encontrada

        Si find_unique es False:
           No lanza errores
           Devuelve lista de resultados
        """

        url = self.addr + "WSRoleServiceService?wsdl"
        client = self.get_client(url)

        roles = client.service.searchRoles(self.s, filtro)

        if find_unique:
            assert (
                len(roles) > 0
            ), f"No se ha encontrado el rol con el filtro: {filtro}. Verifique que sea un filtro LDAP válido."
            assert len(roles) == 1, f"Se ha encontrado más de un rol con: {filtro}"
            return roles[0]
        else:
            return roles

    def lookupRole(self, dn):

        url = self.addr + "WSRoleServiceService?wsdl"
        client = self.get_client(url)

        try:
            r = client.service.lookupRole(self.s, dn)
            return r
        except:
            raise NotFoundError("Rol no encontrado")

    def crearRolEstatico(self, wsrole, wsou):

        url = self.addr + "WSRoleServiceService?wsdl"
        client = self.get_client(url)

        return client.service.createStaticRole(self.s, wsou, wsrole)

    def modificarRolEstatico(self, role_dn, wsattr_list):

        url = self.addr + "WSRoleServiceService?wsdl"
        client = self.get_client(url)

        return client.service.modifyStaticRole(self.s, role_dn, wsattr_list)

    def eliminarRolEstatico(self, role_dn, date=None):

        url = self.addr + "WSRoleServiceService?wsdl"
        client = self.get_client(url)

        if date:
            raise NotImplementedError()
            pass
        else:
            date = Nil
        return client.service.removeRole(self.s, role_dn, date)

    def buscarPersona(self, filtro):

        url = self.addr + "WSPersonServiceService?wsdl"
        client = self.get_client(url)

        personas = client.service.searchPersonsFromRoot(self.s, filtro, Nil)

        assert (
            len(personas) > 0
        ), f"No se ha encontrado la persona con el filtro: {filtro}. Verifique que sea un filtro LDAP válido."
        assert len(personas) == 1, f"Se ha encontrado más de una persona con: {filtro}"
        return personas[0]

    def buscarServicio(self, ou, filtro, find_unique=True):

        url = self.addr + "WSServiceServiceService?wsdl"
        client = self.get_client(url)
        servicios = client.service.searchServices(self.s, ou, filtro)

        if find_unique:
            if len(servicios) == 0:
                raise NotFoundError(
                    f"No se ha encontrado el servicio con el filtro:  {filtro}. Verifique que sea un filtro LDAP válido."
                )
            assert (
                len(servicios) == 1
            ), f"Se ha encontrado más de un servicio con: {filtro}"
            return serialize_object(servicios[0], target_cls=dict)
        else:
            return [serialize_object(s, target_cls=dict) for s in servicios]

    def searchWorkflow(self, nombre, org_name):
        """
        Busca flujos de cuenta y acceso por el nombre.
        Retorna el DN.
        """

        url = self.addr + "WSSearchDataServiceService?wsdl"
        client = self.get_client(url)

        """
        Category puede ser (usar EstaCapitalizacion y quitar _):
        ACCESS_TYPE, ACCOUNT_TEMPLATE, ADOPTION_POLICY, AGENT_OPERATION, ATTRIBUTE_CONSTRAINT, 
        BPUNIT, CATEGORIES_FOR_LIFE_CYCLE_MGT, CONFIG, CONTAINER, CREDENTIAL, CREDENTIAL_COMPONENT, 
        CREDENTIAL_LEASE, CREDENTIAL_POOL, CREDENTIAL_SERVICE, CUSTOM_PROCESS, DYNAMIC_ROLE, FORM_TEMPLATE, 
        GLOBAL_ACCOUNT_TEMPLATE, GROUP, HOST_SELECTION_POLICY, IDENTITY_POLICY, JOIN_DIRECTIVE, JOINDIRECTIVE, 
        LIFECYCLE_PROFILE, LOCATION, OBJECT_PROFILE, ORG, ORGROLE, ORGUNIT, ORPHANED_ACCOUNT, OWNERSHIP_TYPE, 
        PASSWORD_POLICY, PRIVILEGE_RULE, PROVISIONING_POLICY, RECERTIFICATION_POLICY, ROLE, SECURITY_DOMAIN, 
        SEPARATION_OF_DUTY_POLICY, SEPARATION_OF_DUTY_RULE, SERVICE, SERVICE_MODEL, SERVICE_PROFILE, 
        SHARED_ACCESS_POLICY, SYSTEM_ROLE, SYSTEM_USER, TENANT, USERACCESS
        """
        flujos = client.service.findSearchControlObjects(
            self.s,
            {
                "objectclass": "erWorkflowDefinition",
                "contextDN": f"ou=workflow,erglobalid=00000000000000000000,ou={org_name},dc={org_name}",
                "returnedAttributeName": "dn",
                "filter": f"(erProcessName={nombre})",
                "base": "global",
                "category": "CustomProcess",
            },
        )

        assert (
            len(flujos) > 0
        ), f"No se ha encontrado el flujo: {nombre}. Verifique que sea un filtro LDAP válido."
        assert len(flujos) == 1, f"Se ha encontrado más de un servicio con: {nombre}"

        return flujos[0]["value"]

    def buscarGruposPorServicio(self, dn_servicio, profile_name, info):

        url = self.addr + "WSGroupServiceService?wsdl"
        client = self.get_client(url)

        grps = client.service.getGroupsByService(
            self.s, dn_servicio, profile_name, info
        )
        return grps

    def buscarActividadesDeSolicitud(self, process_id):
        """
        getActivities(session: ns1:WSSession, processId: xsd:long, recurseSubProcesses: xsd:boolean) -> getActivitiesReturn: ns1:WSActivity[]
        no funciona bien la búsqueda recursiva de actividades
        """
        url = self.addr + "WSRequestServiceService?wsdl"
        client = self.get_client(url)

        grps = client.service.getActivities(self.s, int(process_id), True)
        return grps

    def suspenderPersona(self,dn,justification):
        #suspendPerson(session: ns1:WSSession, personDN: xsd:string, justification: xsd:string)
        url = self.addr + "WSPersonServiceService?wsdl"
        client = self.get_client(url)

        r = client.service.suspendPerson(self.s, dn, justification)
        return r


    def restaurarPersona(self,dn,justification):
        #restorePerson(password: xsd:string, date: xsd:dateTime, justification: xsd:string) -> restorePersonReturn: ns1:WSRequest
        url = self.addr + "WSPersonServiceService?wsdl"
        client = self.get_client(url)

        r = client.service.restorePerson(self.s, dn, True,Nil,Nil,justification)
        return r

    def eliminarPersona(self,dn,justification):
        #deletePerson(session: ns1:WSSession, personDN: xsd:string, date: xsd:dateTime, justification: xsd:string) -> deletePersonReturn: ns1:WSRequest
        url = self.addr + "WSPersonServiceService?wsdl"
        client = self.get_client(url)

        r = client.service.deletePerson(self.s, dn, Nil, justification)
        return r

    


import requests
from requests.auth import HTTPBasicAuth


class Session:
    """
    Maneja sesión conjunta de REST y SOAP
    """

    def __init__(self, username, password, base_url, certificate_path):
        self.username = username
        # self.password=password
        self.base_url = base_url
        # self.cert=certificate_path

        self.auth = HTTPBasicAuth(username, password)
        self.s = requests.Session()
        self.s.verify = certificate_path if certificate_path else None



# Note: To use the 'upload' functionality of this file, you must:
#   $ pipenv install twine --dev
#Tomado de : https://github.com/navdeep-G/setup.py/blob/master/setup.py

import io
import os
from setuptools import find_packages, setup

# Package meta-data.
NAME = 'pyisim'
DESCRIPTION = "Easy to use Python client for IBM Security Identity Manager (ISIM/ITIM) web services (SOAP and REST APIs) "
URL = 'https://github.com/cazdlt/pyisim'
EMAIL = 'cazdlt@gmail.com'
AUTHOR = 'Andrés Zamora'
REQUIRES_PYTHON = '>=3.8.0'
VERSION = "0.1.1" # Get the version from the package __init__.py
REQUIRED = [
    "requests >= 2.23.0",
    "zeep >= 3.4.0"
]
EXTRAS={}

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Where the magic happens:
setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    download_url=f'{URL}/archive/v{VERSION}.tar.gz',
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*","pycolp"]),
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    keywords = ['isim', 'ibm-security', 'iam',"itim"],
    license='MIT',
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.8',
    ]

)


from pyisim.exceptions import *
from pyisim.entities import (
    Activity,
    Access,
    OrganizationalContainer,
    Person,
    Service,
    StaticRole,
    ProvisioningPolicy,
    Group,
)


def groups(session, by, service_dn=None, group_profile_name="", group_info=""):
    """Buscar grupos.

    Args:
        by (string): "account", "access" or "service"
        dn (string): account or access dn
        groupProfileName (string, optional): Group profile name if searching by service. Defaults to None.
        groupInfo (string, optional): Group name or description if searching by service. Defaults to None.

    Raises:
        NotImplementedError: Search by account or access not implemented yet.
        InvalidOptionError

    Returns:
        list: Search results
    """
    session = session.soapclient
    if by == "account":
        raise NotImplementedError
    elif by == "access":
        raise NotImplementedError
    elif by == "service":
        ret = session.buscarGruposPorServicio(service_dn, group_profile_name, group_info)
    else:
        raise InvalidOptionError("Invalid option")

    return [Group(session, group=g) for g in ret]


def people(
    session,
    filter="*",
    by="cn",
    profile_name="Person",
    attributes="*",
    embedded="",
    limit=50,
):
    """ "Wrapper para buscar/lookup personas o bpperson desde REST API

    Args:
        session ([pyisim.auth.Session]): Sesión de pyisim
        profile (str): Person/BPPerson
        by (str, optional): Atributo por el que buscar. Defaults to "cn".
        filter (str, optional): Filtro por el que buscar. Defaults to "*".
        attributes (str, optional): Atributos a retornar. Defaults to "dn".
        embedded (str, optional): Atributos embebidos a retornar (ej. manager.cn). Defaults to "".
        limit (int, optional): Defaults to 50.

    Returns:
        [list(Person)]: listado de personas encontradas
    """

    ret = session.restclient.buscarPersonas(
        profile_name,
        atributos=attributes,
        embedded=embedded,
        buscar_por=by,
        filtro=filter,
        limit=limit,
    )
    personas = [Person(session, person=p) for p in ret]
    return personas


def provisioning_policy(session, name, parent: OrganizationalContainer):

    wsou = parent.wsou
    results = session.soapclient.buscarPoliticaSuministro(
        wsou, nombre_politica=name, find_unique=False
    )
    return [ProvisioningPolicy(session, provisioning_policy=p) for p in results]


def roles(session, by="errolename", filter="*", find_unique=False):
    soap = session.soapclient
    results = soap.buscarRol(f"({by}={filter})", find_unique)
    return [StaticRole(session, rol=r) for r in results]


def activities(session, by="activityName", filter="*"):
    """Busca actividades

    Args:
        session (pyisim.auth.Session): Sesión de pyisim
        by (str, optional): Filtros disponibles por ISIM REST API (activityId, nombres de actividad/servicio/participantes) o requestId. Defaults to "activityName".
        filter (str, optional): Filtro. Defaults to "*".
    """

    if by == "requestId":
        # session.soapclient.buscarActividadesDeSolicitud(filter)
        """ RESPUESTA IBM A getRecurseSubProcess()
        Your observation is correct. The backend code returns the activities belonging to the first level sub-process only. 
        I feel the main confusion is due to the lack of documentation around this API and the name of the parameter used. 
        The intent of introducing this additional boolean flag was to restrict the result to only the direct activities of the main process. 
        It should have been “getImmediateChildActivities” rather than “recurseSubProcesses”.
        The customer can accomplish this by using a combination of getActivities() and getChildProcesses().
        """
        results = session.restclient.buscarActividad(
            solicitudID=f"/itim/rest/requests/{filter}"
        )
    else:
        results = session.restclient.buscarActividad(
            search_attr=by, search_filter=filter
        )

    return [Activity(session, activity=a) for a in results]


def access(session, by="accessName", filter="*", attributes="", limit=20):

    ret = session.restclient.buscarAcceso(
        by=by, filtro=filter, atributos=attributes, limit=limit
    )
    accesos = [Access(access=a) for a in ret]

    return accesos


def service(session, parent: OrganizationalContainer, by="erservicename", filter="*"):

    # ret=session.restclient.buscarServicio(by,filter,limit,atributos=attributes)
    # servicios=[Service(session,service=s) for s in ret]
    ret = session.soapclient.buscarServicio(
        parent.wsou, f"({by}={filter})", find_unique=False
    )
    servicios = [Service(session, service=s) for s in ret]

    return servicios


def organizational_container(session, profile_name, filter, by="name"):
    """[summary]

    Args:
        session (pyisim.Session)
        profile_name (str): ["bporganizations", "organizationunits", "organizations","locations","admindomains"]
        filter (str): name or attr value
        by (str, optional): attribute to search by. Defaults to "name".

    Returns:
        pyisim.entities.OrganizationalContainer
    """

    buscar_por = None if by == "name" else by
    ret = session.restclient.buscarOUs(
        profile_name, buscar_por=buscar_por, filtro=filter, attributes="dn"
    )

    ous = [OrganizationalContainer(session, organizational_container=ou) for ou in ret]

    return ous
